import wx

from .cmd import WMCmds
from typing import Tuple


class ThreadEvent(wx.PyEvent):
    def __init__(self, id=0) -> None:
        super().__init__(id=id, eventType=wx.wxEVT_THREAD)


class WorkerMessage:
    def __init__(self, cmd: WMCmds, x: int = None, y: int = None,
                 z0: float = None, z1: float = None, z2: float = None) -> None:
        self._cmd = cmd
        self._x = x
        self._y = y
        self._z0 = z0
        self._z1 = z1
        self._z2 = z2

    @property
    def cmd(self) -> WMCmds:
        return self._cmd

    @property
    def x(self) -> int:
        return self._x

    @property
    def y(self) -> int:
        return self._y

    @property
    def z0(self) -> float:
        return self._z0

    @property
    def z1(self) -> float:
        return self._z1

    @property
    def z2(self) -> float:
        return self._z2


def get_bound_box(img: wx.Image, mode: bool, color: wx.Colour) -> Tuple[wx.Size, wx.Point]:
    img_h = img.GetHeight()
    img_w = img.GetWidth()
    bbox = [img_w, -1, 0, 0]
    mask_v = bytearray((color.Red(), color.Green(), color.Blue()))
    data = img.GetData()
    alpha = img.GetAlpha()

    for y in range(img_h):
        has_data = False
        yt = img_w * y
        for x in range(img_w):
            xt = (yt + x) * 3
            if ((not mode and alpha and alpha[yt + x]) or
                    (mode and (data[xt:xt + 3] != mask_v))):
                has_data = True
                if x < bbox[0]:
                    bbox[0] = x
                if x >= bbox[2]:
                    bbox[2] = x + 1

        if has_data:
            if bbox[1] < 0:
                bbox[1] = y
            bbox[3] = y + 1

    if bbox[1] < 0:
        return tuple()

    return (wx.Size(bbox[2] - bbox[0], bbox[3] - bbox[1]),
            wx.Point(-bbox[0], -bbox[1]))


def img_to_heightmap(img: wx.Image, mode: bool, color: wx.Colour) -> wx.Image:
    if mode:
        img.SetMaskColour(color.Red(), color.Green(), color.Blue())

    return wx.Image(img.ConvertToGreyscale())



from Img2STL.utils.cmd import WMCmds
from Img2STL.utils.utils import WorkerMessage
import os
import wx

from threading import Thread
from multiprocessing import Process, Queue
from typing import List

from .cmd import ThrCmds
from .utils import ThreadEvent, get_bound_box, img_to_heightmap
from .worker import Worker
from ..stl import STLFile

from time import sleep


class Executor(Thread):
    def __init__(self,
                 evt_hdlr: wx.EvtHandler,
                 autocrop: bool,
                 density: float,
                 height_max: float,
                 height_min: float,
                 mask_mode: bool,
                 mask_color: wx.Colour,
                 height_mode: bool,
                 f_type: STLFile.BIN or STLFile.ASCII,
                 files_dir: str,
                 files_list: List[str],
                 exit_files_dir: str,
                 exit_files_list: List[str]) -> None:
        super().__init__(name="Executor", daemon=True)
        self._autocrop = autocrop
        self._density = density
        self._height_max = height_max
        self._height_min = height_min
        self._mask_mode = mask_mode
        self._mask_color = mask_color
        self._height_mode = height_mode
        self._f_type = f_type

        self._files_dir = files_dir
        self._files_list = files_list
        self._exit_files_dir = exit_files_dir
        self._exit_files_list = exit_files_list

        self._event = ThreadEvent()
        self._main_evt_handler = evt_hdlr
        self._worker_pool: List[Process] = []
        self._q_to_worker = Queue()
        self._q_from_worker = Queue()

        wx.LogMessage("Executor created.")

    # def __del__(self) -> None:
    #     # self._main_evt_handler._executor = None
    #     wx.LogMessage("Executor destroyed.")

    def _finish(self):
        self._event.SetId(ThrCmds.thrCMD_POOL_END)
        wx.QueueEvent(self._main_evt_handler, self._event.Clone())
        wx.LogMessage("Executor is stopped.")

    def run(self) -> None:
        wx.LogMessage("Executor is running.")

        if len(self._files_list) == 0:
            self._finish()
            return

        points_per_file = 100.0 / len(self._files_list)
        pb_val = 0.0
        cpu_cnt = os.cpu_count()
        if cpu_cnt <= 0:
            cpu_cnt = 1

        mod = self._height_min if self._height_mode else self._height_max
        height_factor = 255.0 / (self._height_max - self._height_min)

        if not self._height_mode:
            height_factor *= -1

        for idx, img_name in enumerate(self._files_list):
            wx.LogMessage(f"Executing for {img_name}")

            self._event.SetId(ThrCmds.thrCMD_POOL_POINT)
            self._event.data = img_name
            wx.QueueEvent(self._main_evt_handler, self._event.Clone())

            for i in range(cpu_cnt):
                proc = Worker(i, self._density, self._q_to_worker, self._q_from_worker)
                self._worker_pool.append(proc)
                proc.start()
                wx.LogMessage(f"{proc.name} is running.")

            if len(self._worker_pool) == 0:
                wx.LogError("Worker Pool is empty!")
                break

            bw = wx.Image(os.path.join(self._files_dir, img_name))

            if self._autocrop:
                bbox = get_bound_box(bw, self._mask_mode, self._mask_color)
                if bbox:
                    bw.Resize(bbox[0], bbox[1])

            bw = img_to_heightmap(bw, self._mask_mode, self._mask_color)

            if not bw.HasAlpha():
                bw.InitAlpha()

            data: bytearray = bw.GetData()
            alpha: bytearray = bw.GetAlpha()
            img_h = bw.GetHeight()
            img_w = bw.GetWidth()
            tot_count = len(alpha)

            for y in range(img_h):
                for x in range(img_w):
                    z0 = z1 = z2 = 0
                    if alpha[(img_w * y) + x]:
                        z0 = data[((img_w * y) + x) * 3] / height_factor + mod
                    if ((x + 1) < img_w) and alpha[(img_w * y) + x + 1]:
                        z1 = data[((img_w * y) + x + 1) * 3] / height_factor + mod
                    if ((y - 1) >= 0) and alpha[(img_w * (y - 1)) + x]:
                        z2 = data[((img_w * (y - 1)) + x) * 3] / height_factor + mod

                    self._q_to_worker.put(
                        WorkerMessage(WMCmds.wMSG_RUN, x, img_h - y - 1, z0, z1, z2))

            for _ in self._worker_pool:
                self._q_to_worker.put(WorkerMessage(WMCmds.wMSG_END))

            while True:
                i = self._q_to_worker.qsize()
                val = pb_val + points_per_file * (1 - (i / tot_count))
                self._event.SetId(ThrCmds.thrCMD_PBAR_UPD)
                self._event.data = int(val)
                wx.QueueEvent(self._main_evt_handler, self._event.Clone())
                if i == 0:
                    pb_val += points_per_file
                    break
                sleep(0.1)

            # FIXME: processes do not terminated
            # for proc in self._worker_pool:
            #     proc.join()
            while self._q_from_worker.qsize() != len(self._worker_pool):
                sleep(0.05)

            for proc in self._worker_pool:
                wx.LogMessage(f"{proc.name} is stopped.")

            self._worker_pool.clear()
            self._generate_stl(idx)

        self._finish()

    def _generate_stl(self, idx: int):
        f_name = self._exit_files_list[idx]
        result_file = STLFile(f_name)

        if self._f_type == STLFile.BIN:
            result_file.set_header(
                f"Generate from Img2STL by baskiton. {f_name}")

        for _ in range(self._q_from_worker.qsize()):
            stl: STLFile = self._q_from_worker.get()
            result_file.add_stl(stl)

        result_file.save_file(self._f_type,
                              os.path.join(self._exit_files_dir, f_name))



import wx
from .ui import Img2STLMainFrame


if __name__ == "__main__":
    app = wx.App(False)
    frame = Img2STLMainFrame()
    frame.SetIcon(frame.app_icon)
    frame.Show(True)
    app.MainLoop()



import wx
import wx.adv

from .ui import AboutDialog


class Img2STLAboutDialog(AboutDialog):
    def __init__(self, parent):
        super().__init__(parent)

    def fit(self) -> None:
        self.Layout()
        self.m_sizer_main.Fit(self)

    def set_icon(self, icon: wx.Icon) -> None:
        self.m_app_icon.SetIcon(icon)

    def set_name(self, name: str, version: str) -> None:
        self.m_name_version.SetLabel(f"{name} {version}")

    def set_description(self, descr: str) -> None:
        self.m_description.SetLabel(descr)

    def set_using(self, using: str) -> None:
        self.m_using_str.SetLabel(using)

    def set_license(self, license: str) -> None:
        pane: wx.Window = self.m_license_pane.GetPane()
        lic = wx.TextCtrl(pane, value=license,
                          style=wx.TE_BESTWRAP | wx.TE_MULTILINE | wx.TE_READONLY | wx.TE_WORDWRAP)

        self.m_sizer_license.Add(lic, 1, wx.EXPAND, 5)

        pane.SetSizer(self.m_sizer_license)
        pane.Layout()
        self.m_sizer_license.Fit(pane)

    def set_copyright(self, url: str, label: str) -> None:
        self.m_copyright.SetURL(url)
        self.m_copyright.SetLabel(label)

    def add_link(self, url: str, label: str) -> None:
        link = wx.adv.HyperlinkCtrl(self, label=label, url=url)
        self.m_url_sizer.Add(link, 0, wx.ALL, 5)



# -*- coding: utf-8 -*-

###########################################################################
## Python code generated with wxFormBuilder (version Oct 26 2018)
## http://www.wxformbuilder.org/
##
## PLEASE DO *NOT* EDIT THIS FILE!
###########################################################################

import wx
import wx.xrc
import wx.aui
import wx.propgrid as pg
import wx.adv

###########################################################################
## Class MainFrame
###########################################################################

class MainFrame ( wx.Frame ):

    def __init__( self, title ):
        wx.Frame.__init__ ( self, None, id = wx.ID_ANY, title = title, pos = wx.DefaultPosition, size = wx.Size( 800,600 ), style = wx.DEFAULT_FRAME_STYLE|wx.TAB_TRAVERSAL )

        self.SetSizeHints( wx.Size( 700,530 ), wx.DefaultSize )
        self.SetFont( wx.Font( 9, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL, False, "Noto Sans" ) )

        self.m_menu_bar = wx.MenuBar( 0 )
        self.m_menu_bar.SetFont( wx.Font( 9, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL, False, "Noto Sans" ) )

        self.m_menu_file = wx.Menu()
        self.m_menu_file.AppendSeparator()

        self.m_menu_file_exit = wx.MenuItem( self.m_menu_file, wx.ID_EXIT, u"&Quit", wx.EmptyString, wx.ITEM_NORMAL )
        self.m_menu_file.Append( self.m_menu_file_exit )

        self.m_menu_bar.Append( self.m_menu_file, u"File" )

        self.m_menu_help = wx.Menu()
        self.m_menu_help.AppendSeparator()

        self.m_menu_help_about = wx.MenuItem( self.m_menu_help, wx.ID_ABOUT, u"&About", wx.EmptyString, wx.ITEM_NORMAL )
        self.m_menu_help.Append( self.m_menu_help_about )

        self.m_menu_bar.Append( self.m_menu_help, u"Help" )

        self.SetMenuBar( self.m_menu_bar )

        m_main_sizer = wx.BoxSizer( wx.VERTICAL )

        self.m_main_container = wx.SplitterWindow( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.SP_LIVE_UPDATE|wx.SP_NOSASH )
        self.m_main_container.Bind( wx.EVT_IDLE, self.m_main_containerOnIdle )

        self.m_prop_panel = wx.Panel( self.m_main_container, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.TAB_TRAVERSAL )
        self.m_prop_panel.SetMinSize( wx.Size( 200,-1 ) )

        m_prop_sizer = wx.BoxSizer( wx.VERTICAL )

        m_prop_image_open_sizer = wx.StaticBoxSizer( wx.StaticBox( self.m_prop_panel, wx.ID_ANY, u"Open Image" ), wx.HORIZONTAL )

        self.m_prop_image_open = wx.Button( m_prop_image_open_sizer.GetStaticBox(), wx.ID_OPEN, u"Open Image(s)", wx.DefaultPosition, wx.DefaultSize, 0 )

        self.m_prop_image_open.SetDefault()
        m_prop_image_open_sizer.Add( self.m_prop_image_open, 1, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.m_prop_image_autocrop = wx.CheckBox( m_prop_image_open_sizer.GetStaticBox(), wx.ID_ANY, u"Auto Crop", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_prop_image_autocrop.SetValue(True)
        m_prop_image_open_sizer.Add( self.m_prop_image_autocrop, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )


        m_prop_sizer.Add( m_prop_image_open_sizer, 0, wx.RIGHT|wx.LEFT|wx.EXPAND, 5 )

        m_prop_prop_sizer = wx.StaticBoxSizer( wx.StaticBox( self.m_prop_panel, wx.ID_ANY, u"Property" ), wx.VERTICAL )

        m_prop_prop_grid = wx.FlexGridSizer( 0, 2, 3, 3 )
        m_prop_prop_grid.SetFlexibleDirection( wx.BOTH )
        m_prop_prop_grid.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )

        self.m_prop_den_lable = wx.StaticText( m_prop_prop_sizer.GetStaticBox(), wx.ID_ANY, u"Density (mm/pix)", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_prop_den_lable.Wrap( -1 )

        m_prop_prop_grid.Add( self.m_prop_den_lable, 0, wx.ALIGN_CENTER_VERTICAL|wx.LEFT, 5 )

        self.m_prop_den_val = wx.SpinCtrlDouble( m_prop_prop_sizer.GetStaticBox(), wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.Size( -1,-1 ), 0, 0, 1000, 2, 0.001 )
        self.m_prop_den_val.SetDigits( 3 )
        m_prop_prop_grid.Add( self.m_prop_den_val, 0, wx.EXPAND|wx.RIGHT|wx.LEFT|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.m_prop_max_height_lable = wx.StaticText( m_prop_prop_sizer.GetStaticBox(), wx.ID_ANY, u"Max Height (mm)", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_prop_max_height_lable.Wrap( -1 )

        m_prop_prop_grid.Add( self.m_prop_max_height_lable, 0, wx.ALIGN_CENTER_VERTICAL|wx.LEFT, 5 )

        self.m_prop_max_height_val = wx.SpinCtrlDouble( m_prop_prop_sizer.GetStaticBox(), wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.Size( -1,-1 ), 0, 0, 1000, 4, 0.001 )
        self.m_prop_max_height_val.SetDigits( 3 )
        m_prop_prop_grid.Add( self.m_prop_max_height_val, 0, wx.RIGHT|wx.LEFT|wx.EXPAND|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.m_prop_min_height_lable = wx.StaticText( m_prop_prop_sizer.GetStaticBox(), wx.ID_ANY, u"Min Height (mm)", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_prop_min_height_lable.Wrap( -1 )

        m_prop_prop_grid.Add( self.m_prop_min_height_lable, 0, wx.ALIGN_CENTER_VERTICAL|wx.LEFT, 5 )

        self.m_prop_min_height_val = wx.SpinCtrlDouble( m_prop_prop_sizer.GetStaticBox(), wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.Size( -1,-1 ), 0, 0, 1000, 1, 0.001 )
        self.m_prop_min_height_val.SetDigits( 3 )
        m_prop_prop_grid.Add( self.m_prop_min_height_val, 0, wx.RIGHT|wx.LEFT|wx.EXPAND|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.m_prop_mask_mode_lable = wx.StaticText( m_prop_prop_sizer.GetStaticBox(), wx.ID_ANY, u"Mask Mode", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_prop_mask_mode_lable.Wrap( -1 )

        m_prop_prop_grid.Add( self.m_prop_mask_mode_lable, 0, wx.ALIGN_CENTER_VERTICAL|wx.LEFT, 5 )

        m_prop_mask_mode_valChoices = [ u"Transparency", u"Color" ]
        self.m_prop_mask_mode_val = wx.Choice( m_prop_prop_sizer.GetStaticBox(), wx.ID_ANY, wx.DefaultPosition, wx.Size( -1,-1 ), m_prop_mask_mode_valChoices, 0 )
        self.m_prop_mask_mode_val.SetSelection( 0 )
        m_prop_prop_grid.Add( self.m_prop_mask_mode_val, 0, wx.RIGHT|wx.LEFT|wx.EXPAND|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.m_prop_transp_col_lable = wx.StaticText( m_prop_prop_sizer.GetStaticBox(), wx.ID_ANY, u"Transparency color", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_prop_transp_col_lable.Wrap( -1 )

        self.m_prop_transp_col_lable.Enable( False )

        m_prop_prop_grid.Add( self.m_prop_transp_col_lable, 0, wx.ALIGN_CENTER_VERTICAL|wx.LEFT, 5 )

        self.m_prop_transp_col_val = wx.Button( m_prop_prop_sizer.GetStaticBox(), wx.ID_ANY, u"#FF00FF", wx.DefaultPosition, wx.DefaultSize, wx.BORDER_NONE )
        self.m_prop_transp_col_val.SetBackgroundColour( wx.Colour( 255, 0, 255 ) )
        self.m_prop_transp_col_val.Enable( False )

        m_prop_prop_grid.Add( self.m_prop_transp_col_val, 0, wx.RIGHT|wx.LEFT|wx.ALIGN_CENTER_VERTICAL|wx.EXPAND, 5 )

        self.m_prop_height_mode_lable = wx.StaticText( m_prop_prop_sizer.GetStaticBox(), wx.ID_ANY, u"Height Mode", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_prop_height_mode_lable.Wrap( -1 )

        m_prop_prop_grid.Add( self.m_prop_height_mode_lable, 0, wx.ALIGN_CENTER_VERTICAL|wx.LEFT, 5 )

        m_prop_height_mode_valChoices = [ u"Black", u"White" ]
        self.m_prop_height_mode_val = wx.Choice( m_prop_prop_sizer.GetStaticBox(), wx.ID_ANY, wx.DefaultPosition, wx.Size( -1,-1 ), m_prop_height_mode_valChoices, 0 )
        self.m_prop_height_mode_val.SetSelection( 1 )
        m_prop_prop_grid.Add( self.m_prop_height_mode_val, 0, wx.EXPAND|wx.ALIGN_CENTER_VERTICAL|wx.BOTTOM|wx.RIGHT|wx.LEFT, 5 )


        m_prop_prop_sizer.Add( m_prop_prop_grid, 0, 0, 5 )


        m_prop_sizer.Add( m_prop_prop_sizer, 0, wx.EXPAND|wx.RIGHT|wx.LEFT, 5 )

        m_prop_exit_sizer = wx.StaticBoxSizer( wx.StaticBox( self.m_prop_panel, wx.ID_ANY, u"Exit File" ), wx.VERTICAL )

        m_prop_exit_grid = wx.BoxSizer( wx.HORIZONTAL )

        self.m_prop_exit_label = wx.StaticText( m_prop_exit_sizer.GetStaticBox(), wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, wx.ALIGN_LEFT|wx.ST_ELLIPSIZE_MIDDLE|wx.ST_NO_AUTORESIZE )
        self.m_prop_exit_label.Wrap( -1 )

        m_prop_exit_grid.Add( self.m_prop_exit_label, 1, wx.ALIGN_CENTER_VERTICAL|wx.RIGHT|wx.LEFT, 5 )

        self.m_prop_exit_btn = wx.Button( m_prop_exit_sizer.GetStaticBox(), wx.ID_ANY, u"Choose", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_prop_exit_btn.Enable( False )

        m_prop_exit_grid.Add( self.m_prop_exit_btn, 0, wx.ALIGN_CENTER_VERTICAL|wx.LEFT, 5 )


        m_prop_exit_sizer.Add( m_prop_exit_grid, 0, wx.EXPAND|wx.BOTTOM|wx.RIGHT|wx.LEFT, 5 )

        m_prop_exit_formatChoices = [ u"ASCII", u"BIN" ]
        self.m_prop_exit_format = wx.RadioBox( m_prop_exit_sizer.GetStaticBox(), wx.ID_ANY, u"STL format", wx.DefaultPosition, wx.DefaultSize, m_prop_exit_formatChoices, 1, wx.RA_SPECIFY_ROWS )
        self.m_prop_exit_format.SetSelection( 1 )
        m_prop_exit_sizer.Add( self.m_prop_exit_format, 0, wx.BOTTOM|wx.LEFT, 5 )


        m_prop_sizer.Add( m_prop_exit_sizer, 0, wx.EXPAND|wx.RIGHT|wx.LEFT, 5 )

        m_prop_execute_sizer = wx.StaticBoxSizer( wx.StaticBox( self.m_prop_panel, wx.ID_ANY, u"Execute" ), wx.VERTICAL )

        self.m_execute_button = wx.Button( m_prop_execute_sizer.GetStaticBox(), wx.ID_ANY, u"Execute", wx.DefaultPosition, wx.DefaultSize, 0 )

        self.m_execute_button.SetDefault()
        self.m_execute_button.Enable( False )

        m_prop_execute_sizer.Add( self.m_execute_button, 0, wx.ALIGN_CENTER_HORIZONTAL, 5 )

        self.m_progress_bar = wx.Gauge( m_prop_execute_sizer.GetStaticBox(), wx.ID_ANY, 100, wx.DefaultPosition, wx.DefaultSize, wx.GA_HORIZONTAL|wx.GA_SMOOTH )
        self.m_progress_bar.SetValue( 0 )
        m_prop_execute_sizer.Add( self.m_progress_bar, 0, wx.ALL|wx.EXPAND, 5 )

        self.m_progress_label = wx.StaticText( m_prop_execute_sizer.GetStaticBox(), wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_progress_label.Wrap( -1 )

        m_prop_execute_sizer.Add( self.m_progress_label, 0, wx.ALIGN_CENTER_HORIZONTAL|wx.BOTTOM, 5 )


        m_prop_sizer.Add( m_prop_execute_sizer, 0, wx.EXPAND|wx.BOTTOM|wx.RIGHT|wx.LEFT, 5 )


        self.m_prop_panel.SetSizer( m_prop_sizer )
        self.m_prop_panel.Layout()
        m_prop_sizer.Fit( self.m_prop_panel )
        self.m_imagepanel = wx.ScrolledWindow( self.m_main_container, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.HSCROLL|wx.VSCROLL )
        self.m_imagepanel.SetScrollRate( 5, 5 )
        m_image_panel_sizer = wx.BoxSizer( wx.VERTICAL )

        self.m_image_panel_container = wx.SplitterWindow( self.m_imagepanel, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.SP_LIVE_UPDATE|wx.SP_NOSASH )
        self.m_image_container = wx.ScrolledWindow( self.m_image_panel_container, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.BORDER_SUNKEN|wx.HSCROLL|wx.VSCROLL )
        self.m_image_container.SetScrollRate( 5, 5 )
        m_image_sizer = wx.BoxSizer( wx.VERTICAL )

        self.m_note = wx.aui.AuiNotebook( self.m_image_container, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.aui.AUI_NB_CLOSE_ON_ALL_TABS|wx.aui.AUI_NB_SCROLL_BUTTONS|wx.aui.AUI_NB_TAB_MOVE|wx.aui.AUI_NB_TAB_SPLIT|wx.aui.AUI_NB_TOP )

        m_image_sizer.Add( self.m_note, 1, wx.EXPAND, 5 )


        self.m_image_container.SetSizer( m_image_sizer )
        self.m_image_container.Layout()
        m_image_sizer.Fit( self.m_image_container )
        self.m_image_info_panel = wx.ScrolledWindow( self.m_image_panel_container, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_image_info_panel.SetScrollRate( 5, 5 )
        self.m_image_info_panel.SetMinSize( wx.Size( 200,-1 ) )

        m_image_info_panel_sizer = wx.BoxSizer( wx.VERTICAL )

        m_palette_sizer = wx.StaticBoxSizer( wx.StaticBox( self.m_image_info_panel, wx.ID_ANY, u"Palette" ), wx.VERTICAL )

        self.m_palette_grid = pg.PropertyGrid(m_palette_sizer.GetStaticBox(), wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.propgrid.PG_HIDE_MARGIN|wx.propgrid.PG_LIMITED_EDITING)
        self.m_palette_grid.SetMaxSize( wx.Size( 200,-1 ) )

        self.m_palette_header = self.m_palette_grid.Append( pg.PropertyCategory( u"amout | color", u"amout | color" ) )
        m_palette_sizer.Add( self.m_palette_grid, 1, wx.EXPAND, 5 )


        m_image_info_panel_sizer.Add( m_palette_sizer, 1, wx.EXPAND, 5 )


        self.m_image_info_panel.SetSizer( m_image_info_panel_sizer )
        self.m_image_info_panel.Layout()
        m_image_info_panel_sizer.Fit( self.m_image_info_panel )
        self.m_image_panel_container.SplitVertically( self.m_image_container, self.m_image_info_panel, -1 )
        m_image_panel_sizer.Add( self.m_image_panel_container, 1, wx.EXPAND, 5 )


        self.m_imagepanel.SetSizer( m_image_panel_sizer )
        self.m_imagepanel.Layout()
        m_image_panel_sizer.Fit( self.m_imagepanel )
        self.m_main_container.SplitVertically( self.m_prop_panel, self.m_imagepanel, 280 )
        m_main_sizer.Add( self.m_main_container, 1, wx.EXPAND, 5 )


        self.SetSizer( m_main_sizer )
        self.Layout()
        self.m_status_bar = self.CreateStatusBar( 1, wx.STB_SIZEGRIP, wx.ID_ANY )

        self.Centre( wx.BOTH )

        # Connect Events
        self.Bind( wx.EVT_CLOSE, self.OnClose )
        self.Bind( wx.EVT_MENU, self.OnExit, id = self.m_menu_file_exit.GetId() )
        self.Bind( wx.EVT_MENU, self.OnAbout, id = self.m_menu_help_about.GetId() )
        self.m_main_container.Bind( wx.EVT_SPLITTER_SASH_POS_CHANGING, self.main_containerOnSplitterSashPosChanging )
        self.m_prop_image_open.Bind( wx.EVT_BUTTON, self.prop_image_openOnButtonClick )
        self.m_prop_image_autocrop.Bind( wx.EVT_CHECKBOX, self.prop_image_autocropOnCheckBox )
        self.m_prop_mask_mode_val.Bind( wx.EVT_CHOICE, self.prop_mask_mode_valOnChoice )
        self.m_prop_transp_col_val.Bind( wx.EVT_BUTTON, self.prop_transp_col_valOnButtonClick )
        self.m_prop_exit_btn.Bind( wx.EVT_BUTTON, self.prop_exit_btnOnButtonClick )
        self.m_execute_button.Bind( wx.EVT_BUTTON, self.execute_buttonOnButtonClick )
        self.m_image_panel_container.Bind( wx.EVT_SIZE, self.image_panel_containerOnSize )
        self.m_image_panel_container.Bind( wx.EVT_SPLITTER_SASH_POS_CHANGED, self.image_panel_containerOnSplitterSashPosChanged )
        self.m_note.Bind( wx.aui.EVT_AUINOTEBOOK_PAGE_CHANGED, self.noteOnAuiNotebookPageChanged )
        self.m_note.Bind( wx.aui.EVT_AUINOTEBOOK_PAGE_CLOSE, self.noteOnAuiNotebookPageClose )

    def __del__( self ):
        pass


    # Virtual event handlers, overide them in your derived class
    def OnClose( self, event ):
        event.Skip()

    def OnExit( self, event ):
        event.Skip()

    def OnAbout( self, event ):
        event.Skip()

    def main_containerOnSplitterSashPosChanging( self, event ):
        event.Skip()

    def prop_image_openOnButtonClick( self, event ):
        event.Skip()

    def prop_image_autocropOnCheckBox( self, event ):
        event.Skip()

    def prop_mask_mode_valOnChoice( self, event ):
        event.Skip()

    def prop_transp_col_valOnButtonClick( self, event ):
        event.Skip()

    def prop_exit_btnOnButtonClick( self, event ):
        event.Skip()

    def execute_buttonOnButtonClick( self, event ):
        event.Skip()

    def image_panel_containerOnSize( self, event ):
        event.Skip()

    def image_panel_containerOnSplitterSashPosChanged( self, event ):
        event.Skip()

    def noteOnAuiNotebookPageChanged( self, event ):
        event.Skip()

    def noteOnAuiNotebookPageClose( self, event ):
        event.Skip()

    def m_main_containerOnIdle( self, event ):
    	self.m_main_container.SetSashPosition( 280 )
    	self.m_main_container.Unbind( wx.EVT_IDLE )


###########################################################################
## Class AboutDialog
###########################################################################

class AboutDialog ( wx.Dialog ):

    def __init__( self, parent ):
        wx.Dialog.__init__ ( self, parent, id = wx.ID_ANY, title = u"About", pos = wx.DefaultPosition, size = wx.Size( -1,-1 ), style = wx.DEFAULT_DIALOG_STYLE )

        self.SetSizeHints( wx.Size( -1,-1 ), wx.Size( -1,-1 ) )

        self.m_sizer_main = wx.BoxSizer( wx.VERTICAL )

        bSizerInMain = wx.BoxSizer( wx.HORIZONTAL )

        self.m_app_icon = wx.StaticBitmap( self, wx.ID_ANY, wx.NullBitmap, wx.DefaultPosition, wx.Size( 32,32 ), 0 )
        bSizerInMain.Add( self.m_app_icon, 0, wx.ALL, 5 )

        bSizerDesc = wx.BoxSizer( wx.VERTICAL )

        bSizerName = wx.BoxSizer( wx.HORIZONTAL )


        bSizerName.Add( ( 0, 0), 1, wx.EXPAND, 5 )

        self.m_name_version = wx.StaticText( self, wx.ID_ANY, u"%name% %version%", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_name_version.Wrap( -1 )

        self.m_name_version.SetFont( wx.Font( 12, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_BOLD, False, wx.EmptyString ) )

        bSizerName.Add( self.m_name_version, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )


        bSizerName.Add( ( 0, 0), 1, wx.EXPAND, 5 )


        bSizerDesc.Add( bSizerName, 1, wx.EXPAND, 5 )

        self.m_description = wx.StaticText( self, wx.ID_ANY, u"%description%", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_description.Wrap( -1 )

        bSizerDesc.Add( self.m_description, 0, wx.ALL|wx.EXPAND, 5 )

        self.m_using_str = wx.StaticText( self, wx.ID_ANY, u"Using %wx%", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.m_using_str.Wrap( -1 )

        bSizerDesc.Add( self.m_using_str, 0, wx.EXPAND|wx.ALL, 5 )

        self.m_url_sizer = wx.BoxSizer( wx.VERTICAL )


        bSizerDesc.Add( self.m_url_sizer, 1, wx.EXPAND, 5 )


        bSizerInMain.Add( bSizerDesc, 1, wx.EXPAND, 5 )


        self.m_sizer_main.Add( bSizerInMain, 0, wx.EXPAND, 5 )

        self.m_license_pane = wx.CollapsiblePane( self, wx.ID_ANY, u"License", wx.DefaultPosition, wx.DefaultSize, wx.CP_DEFAULT_STYLE )
        self.m_license_pane.Collapse( True )

        self.m_sizer_license = wx.BoxSizer( wx.VERTICAL )


        self.m_license_pane.GetPane().SetSizer( self.m_sizer_license )
        self.m_license_pane.GetPane().Layout()
        self.m_sizer_license.Fit( self.m_license_pane.GetPane() )
        self.m_sizer_main.Add( self.m_license_pane, 1, wx.EXPAND|wx.ALL, 5 )

        bSizerLow = wx.BoxSizer( wx.HORIZONTAL )

        self.m_copyright = wx.adv.HyperlinkCtrl( self, wx.ID_ANY, u"%label%", u"%url%", wx.DefaultPosition, wx.DefaultSize, wx.adv.HL_DEFAULT_STYLE )
        bSizerLow.Add( self.m_copyright, 1, wx.ALL|wx.ALIGN_BOTTOM, 5 )

        m_sdbSizer = wx.StdDialogButtonSizer()
        self.m_sdbSizerOK = wx.Button( self, wx.ID_OK )
        m_sdbSizer.AddButton( self.m_sdbSizerOK )
        m_sdbSizer.Realize();

        bSizerLow.Add( m_sdbSizer, 0, wx.EXPAND|wx.ALL, 5 )


        self.m_sizer_main.Add( bSizerLow, 0, wx.EXPAND, 5 )


        self.SetSizer( self.m_sizer_main )
        self.Layout()
        self.m_sizer_main.Fit( self )

        self.Centre( wx.BOTH )

    def __del__( self ):
        pass





import wx

from multiprocessing import Process, Queue
from ..stl import Vertex3, Polygon3, STLFile
from .utils import WorkerMessage, WMCmds


class Worker(Process):
    def __init__(self, tid: int, den: float, q_from_exec: Queue, q_to_exec: Queue) -> None:
        super().__init__(name=f"Worker-{tid}", daemon=True)
        self._density = den
        self._q_from_exec = q_from_exec
        self._q_to_exec = q_to_exec
        self._file = STLFile()

        wx.LogMessage(f"{self.name} created.")

    def run(self) -> None:

        while True:
            msg: WorkerMessage = self._q_from_exec.get()

            if msg.cmd == WMCmds.wMSG_RUN:
                self._pixel_calc(msg)
            elif msg.cmd == WMCmds.wMSG_END:
                self._q_to_exec.put(self._file)
                break
            elif msg.cmd == WMCmds.wMSG_STOP:
                break

    def _pixel_calc(self, msg: WorkerMessage) -> None:
        if msg.z0 != 0:
            vx_1 = Vertex3(msg.x * self._density,
                           msg.y * self._density,
                           msg.z0)
            vx_2 = Vertex3((msg.x + 1) * self._density,
                           msg.y * self._density,
                           msg.z0)
            vx_3 = Vertex3(msg.x * self._density,
                           (msg.y + 1) * self._density,
                           msg.z0)
            vx_4 = Vertex3((msg.x + 1) * self._density,
                           (msg.y + 1) * self._density,
                           msg.z0)
            nrm = Vertex3(0.0, 0.0, 1.0)

            self._file.add_triangle(Polygon3(nrm, vx_3, vx_2, vx_4))
            self._file.add_triangle(Polygon3(nrm, vx_1, vx_2, vx_3))

            vx_1 = Vertex3(msg.x * self._density,
                           msg.y * self._density,
                           0.0)
            vx_2 = Vertex3((msg.x + 1) * self._density,
                           msg.y * self._density,
                           0.0)
            vx_3 = Vertex3(msg.x * self._density,
                           (msg.y + 1) * self._density,
                           0.0)
            vx_4 = Vertex3((msg.x + 1) * self._density,
                           (msg.y + 1) * self._density,
                           0.0)
            nrm = Vertex3(0.0, 0.0, -1.0)

            self._file.add_triangle(Polygon3(nrm, vx_1, vx_3, vx_2))
            self._file.add_triangle(Polygon3(nrm, vx_3, vx_4, vx_2))

        if msg.z1 != msg.z0:
            vx_1 = Vertex3((msg.x + 1) * self._density,
                           msg.y * self._density,
                           msg.z0)
            vx_2 = Vertex3((msg.x + 1) * self._density,
                           msg.y * self._density,
                           msg.z1)
            vx_3 = Vertex3((msg.x + 1) * self._density,
                           (msg.y + 1) * self._density,
                           msg.z0)
            vx_4 = Vertex3((msg.x + 1) * self._density,
                           (msg.y + 1) * self._density,
                           msg.z1)
            nrm = Vertex3(1.0, 0.0, 0.0)
            if msg.z1 >= msg.z0:
                nrm.x = -1.0

            self._file.add_triangle(Polygon3(nrm, vx_1, vx_2, vx_3))
            self._file.add_triangle(Polygon3(nrm, vx_2, vx_4, vx_3))

        if (msg.x == 0) and (msg.z0 > 0):
            vx_1 = Vertex3(msg.x * self._density,
                           msg.y * self._density,
                           msg.z0)
            vx_2 = Vertex3(msg.x * self._density,
                           (msg.y + 1) * self._density,
                           msg.z0)
            vx_3 = Vertex3(msg.x * self._density,
                           (msg.y + 1) * self._density,
                           0.0)
            vx_4 = Vertex3(msg.x * self._density,
                           msg.y * self._density,
                           0.0)
            nrm = Vertex3(-1.0, 0.0, 0.0)

            self._file.add_triangle(Polygon3(nrm, vx_1, vx_2, vx_3))
            self._file.add_triangle(Polygon3(nrm, vx_1, vx_3, vx_4))

        if msg.z2 != msg.z0:
            vx_1 = Vertex3(msg.x * self._density,
                           (msg.y + 1) * self._density,
                           msg.z0)
            vx_2 = Vertex3((msg.x + 1) * self._density,
                           (msg.y + 1) * self._density,
                           msg.z2)
            vx_3 = Vertex3(msg.x * self._density,
                           (msg.y + 1) * self._density,
                           msg.z2)
            vx_4 = Vertex3((msg.x + 1) * self._density,
                           (msg.y + 1) * self._density,
                           msg.z0)
            nrm = Vertex3(0.0, 1.0, 0.0)
            if msg.z2 >= msg.z0:
                nrm.y = -1.0

            self._file.add_triangle(Polygon3(nrm, vx_1, vx_2, vx_3))
            self._file.add_triangle(Polygon3(nrm, vx_1, vx_4, vx_2))

        if (msg.y == 0) and (msg.z0 > 0):
            vx_1 = Vertex3(msg.x * self._density,
                           msg.y * self._density,
                           msg.z0)
            vx_2 = Vertex3(msg.x * self._density,
                           msg.y * self._density,
                           0.0)
            vx_3 = Vertex3((msg.x + 1) * self._density,
                           msg.y * self._density,
                           msg.z0)
            vx_4 = Vertex3((msg.x + 1) * self._density,
                           msg.y * self._density,
                           0.0)
            nrm = Vertex3(0.0, -1.0, 0.0)

            self._file.add_triangle(Polygon3(nrm, vx_1, vx_2, vx_3))
            self._file.add_triangle(Polygon3(nrm, vx_2, vx_4, vx_3))



import os
import wx

from datetime import datetime


class Logging(wx.Log):
    def __init__(self, log_file_name: str) -> None:
        super().__init__()
        self._old_logger: wx.Log = wx.Log.GetActiveTarget()
        self._log_file_name = os.path.join(os.getcwd(),  log_file_name)

        wx.Log.SetActiveTarget(self)

        with open(self._log_file_name, "w+") as flog:
            flog.write(f"Session started at {wx.DateTime.Now().FormatISODate()}\n")

    def __del__(self) -> None:
        wx.Log.SetActiveTarget(self._old_logger)

    def do_log_line(self, level: str, timestr: str, msg: str) -> None:
        with open(self._log_file_name, "a+") as flog:
            t_str = f"<{level:8}> {timestr:8} {msg}\n"
            flog.write(t_str)

    def DoLogRecord(self, level, msg: str, info: wx.LogRecordInfo) -> None:
        postfix = ""

        if level <= wx.LOG_Warning or level == wx.LOG_Trace or level == wx.LOG_Debug:
            postfix = f"[{info.func}() in {info.filename}({info.line})]"
            self._old_logger.LogRecord(level, f"{level}: {msg}", info)

        if level == wx.LOG_Error:
            level_str = "ERROR"
        elif level == wx.LOG_Warning:
            level_str = "WARNING"
        elif level == wx.LOG_Message:
            level_str = "MESSAGE"
        elif level == wx.LOG_Status:
            level_str = "STATUS"
        elif level == wx.LOG_Info:
            level_str = "INFO"
        elif level == wx.LOG_Debug: 
            level_str = "DEBUG"
        elif level == wx.LOG_Trace:
            level_str = "TRACE"
        elif level == wx.LOG_Progress:
            level_str = "PROGRESS"
        else:
            level_str = "USER"

        self.do_log_line(
            level_str,
            datetime.utcfromtimestamp(info.timestamp).strftime("%H:%M:%S"),
            f"{msg} {postfix}"
        )



import pkg_resources

from .. import __version__


APP_ICON = pkg_resources.resource_filename(__name__, 'app_icon.ico')

ST_NAME = "Image to STL"
ST_VERSION = __version__
ST_DESCRIPTION = "Application for converting images to STL (3D model for printing)"
ST_COPYRIGHT = "Copyright (c) 2019-2021 Alexander Baskikh"
ST_LICENSE_LINK = "https://github.com/baskiton/Img2STL/blob/main++/LICENSE"
ST_LICENSE =    \
        "MIT License\n\n"       \
        "Copyright (c) 2019-2021 Alexander Baskikh\n\n" \
        "Permission is hereby granted, free of charge, to any person obtaining a copy " \
        "of this software and associated documentation files (the \"Software\"), to deal "      \
        "in the Software without restriction, including without limitation the rights to "      \
        "use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies "        \
        "of the Software, and to permit persons to whom the Software is furnished to do "       \
        "so, subject to the following conditions:\n\n"  \
        "The above copyright notice and this permission notice shall be included in all "       \
        "copies or substantial portions of the Software.\n\n"   \
        "THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR " \
        "IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, "     \
        "FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE "  \
        "AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, "    \
        "WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN "      \
        "CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE."



from __future__ import annotations

import struct


class Vertex3:
    def __init__(self, x: float, y: float, z: float):
        self.x = x
        self.y = y
        self.z = z

    def pack(self) -> bytes:
        return struct.pack("3f", self.x, self.y, self.z)

    def __str__(self) -> str:
        return f"{self.x:e} {self.y:e} {self.z:e}"

    def __repr__(self) -> str:
        return f"Vertex3({self.x:f}, {self.y:f}, {self.z:f})"


class Polygon3():
    def __init__(self, nrm, vx_1: Vertex3, vx_2: Vertex3, vx_3: Vertex3):
        self.nrm = nrm
        self.vx_1 = vx_1
        self.vx_2 = vx_2
        self.vx_3 = vx_3

    def to_bytes(self) -> bytes:
        return self.nrm.pack() + self.vx_1.pack() + self.vx_2.pack() + self.vx_3.pack() + b"\0\0"

    def __repr__(self) -> str:
        return f"Polygon3(nrm=[{self.nrm}], x=[{self.vx_1}], y=[{self.vx_2}], z=[{self.vx_3}])"



from __future__ import annotations

import os
import wx
import struct

from typing import List
from .geometry import Polygon3


class STLFile:
    ASCII = 0
    BIN = 1

    def __init__(self, f_name: str = "") -> None:
        self._name = f"{f_name:80}"
        if len(self._name) > 80:
            self._name = self._name[:80]
        self._polygons_list: List[Polygon3] = []

    def add_triangle(self, polygon: Polygon3) -> None:
        self._polygons_list.append(polygon)

    def add_stl(self, stl: STLFile) -> None:
        self._polygons_list.extend(stl._polygons_list)

    def clear(self) -> None:
        self._polygons_list.clear()

    def set_header(self, hdr: str) -> None:
        self._name = f"{hdr:80}"
        if len(self._name) > 80:
            self._name = self._name[:80]

    def save_file(self, f_type, f_name: str) -> bool:
        mode = "w" if f_type == STLFile.ASCII else "wb"

        with open(f_name, mode) as file:
            result = self.to_ascii() if f_type == STLFile.ASCII else self.to_bin()
            file.write(result)

        wx.LogMessage(f"File {os.path.basename(f_name)} created.")

        return True

    def polygons(self) -> int:
        return len(self._polygons_list)

    def to_ascii(self) -> str:
        result = "solid {}\n".format(self._name.replace(" \n\t", '_'))

        for polygon in self._polygons_list:
            result += f" facet normal {polygon.nrm}\n"  \
                      f"  outer loop\n"                 \
                      f"   vertex {polygon.vx_1}\n"     \
                      f"   vertex {polygon.vx_2}\n"     \
                      f"   vertex {polygon.vx_3}\n"     \
                      f"  endloop\n endfacet\n"

        result += f"endsolid {self._name}"

        return result

    def to_bin(self) -> bytearray:
        result = bytearray()

        result += struct.pack("80s", bytes(self._name, "ascii"))    # header
        result += struct.pack("I", self.polygons())
        for polygon in self._polygons_list:
            result += polygon.to_bytes()

        return result



import pathlib
import sys

from setuptools import setup, find_packages

# Check python version
MINIMAL_PY_VERSION = (3, 7)
if sys.version_info < MINIMAL_PY_VERSION:
    raise RuntimeError('This app works only with Python {}+'.format('.'.join(map(str, MINIMAL_PY_VERSION))))


def get_version():
    here = pathlib.Path(__file__).parent
    file =(here / 'Img2STL' / '__init__.py').read_text('utf-8')
    for line in file.splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]


def get_long_description():
    with open("README.md", "r", encoding="utf-8") as readme:
        return readme.read()


setup(
    name='Img2STL',
    version=get_version(),
    url='https://github.com/baskiton/Img2STL',
    project_urls={
        "Bug Tracker": "https://github.com/baskiton/Img2STL/issues",
    },
    license='MIT',
    author='Alexander Baskikh',
    python_requires='>=3.7',
    author_email='baskiton@gmail.com',
    description='Application for converting images to STL (3D model for printing)',
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    package_data={
        '':['app_icon.ico'],
    },
    install_requires=[
        'numpy==1.20.3',
        'Pillow==8.2.0',
        'six==1.16.0',
        'wxPython==4.1.1;platform_system!="Linux"',
    ],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Multimedia :: Graphics :: 3D Modeling',
    ]
)



import os
import wx
import wx.propgrid as wx_pg

from typing import List

from .ui import MainFrame
from .about_dialog import Img2STLAboutDialog
from ..static import *
from ..utils import get_bound_box, ThrCmds, ThreadEvent, Logging, Executor


def wxbuildinfo():
    wx_build = wx.VERSION_STRING
    platform = wx.PlatformInformation()
    os_name = platform.GetOperatingSystemFamilyName()

    return f"wxPython-{wx_build}-{os_name}"


class Img2STLMainFrame(MainFrame):
    def __init__(self, title: str = "Img2STL"):
        super().__init__(title=title)

        self._files_directory = os.getcwd()
        self._exit_files_dir = os.getcwd()
        self._files_list: List[str] = []
        self._exit_files: List[str] = []

        self._logger = Logging("log.txt")
        self._executor: Executor or None = None

        self._cnt = False
        self.app_icon = wx.Icon(APP_ICON)

        self.Fit()

        self.Bind(wx.EVT_THREAD, self.on_thread_update)
        wx.LogMessage("The program is running.")

    def on_thread_update(self, event: ThreadEvent) -> None:
        ev_id: ThrCmds = event.GetId()

        if ev_id == ThrCmds.thrCMD_POOL_END:
            self.m_progress_label.SetLabel("Done!")
            self.m_progress_label.CenterOnParent(wx.HORIZONTAL)
            self.m_prop_panel.Enable(True)
            self.m_progress_bar.SetValue(100)
            self._executor.join()
            del self._executor
            self._executor = None

        elif ev_id == ThrCmds.thrCMD_POOL_POINT:
            self.m_progress_label.SetLabel(event.data)
            self.m_progress_label.CenterOnParent(wx.HORIZONTAL)

        elif ev_id == ThrCmds.thrCMD_PBAR_UPD:
            self.m_progress_bar.SetValue(event.data)

    def set_prop_exit_label(self) -> None:
        fn = self._exit_files_dir
        cnt = len(self._files_list)

        if cnt == 0:
            return
        elif cnt == 1:
            fn = os.path.join(fn, self._exit_files[0])

        self.m_prop_exit_label.SetLabel(fn)
        self.m_prop_exit_label.SetToolTip(fn)

    def add_image_tabs(self) -> None:
        if len(self._files_list) == 0:
            return

        self.m_note.DeleteAllPages()

        for item in self._files_list:
            tab = wx.Panel(self)
            self.m_note.AddPage(tab, item)

    def change_image_scale(self) -> None:
        if len(self._files_list) == 0:
            return

        cont_size: wx.Size = self.m_image_container.GetSize()
        cont_ratio = cont_size.GetWidth() / cont_size.GetHeight()

        fn = os.path.join(self._files_directory, self.m_note.GetPageText(self.m_note.GetSelection()))

        img = wx.Image(fn)
        if not img.IsOk():
            return

        if self.m_prop_image_autocrop.GetValue():
            bbox = get_bound_box(img, self.m_prop_mask_mode_val.GetSelection(),
                                 self.m_prop_transp_col_val.GetBackgroundColour())

            if bbox:
                img.Resize(bbox[0], bbox[1])

        if self.m_prop_mask_mode_val.GetSelection():
            col: wx.Colour = self.m_prop_transp_col_val.GetBackgroundColour()
            img.SetMaskColour(col.Red(), col.Green(), col.Blue())

        if not img.HasAlpha():
            img.InitAlpha()

        self.fill_palette(img)

        img_ratio = img.GetWidth() / img.GetHeight()

        if cont_ratio > img_ratio:
            w = (img.GetWidth() * cont_size.GetHeight()) / img.GetHeight() * 0.8
            h = cont_size.GetHeight() * 0.8
        else:
            w = cont_size.GetWidth() * 0.8
            h = (img.GetHeight() * cont_size.GetWidth()) / img.GetWidth() * 0.8

        img.Rescale(w, h)

        tab: wx.Panel = self.m_note.GetCurrentPage()
        tab.DestroyChildren()

        bm = wx.StaticBitmap(tab, bitmap=wx.Bitmap(img))
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(bm, 1, wx.ALL | wx.EXPAND, 5)

        del img

        tab.SetSizer(sizer)
        tab.Layout()
        sizer.Fit(tab)
        self.m_note.Layout()

    def fill_palette(self, img: wx.Image) -> None:
        table = dict()
        data: bytearray = img.GetData()
        alpha: bytearray = img.GetAlpha()

        for i in range(img.GetHeight() * img.GetWidth()):
            if alpha[i] == 0:
                continue
            idx = i*3
            r = data[idx]
            g = data[idx+1]
            b = data[idx+2]
            rgb = (r, g, b)
            if rgb in table:
                table[rgb][0] += 1
            else:
                table[rgb] = [1, wx.Colour(r, g, b)]

        self.m_palette_grid.Clear()
        self.m_palette_grid.Append(wx_pg.PropertyCategory("amount | color"))

        for i in table.values():
            self.m_palette_grid.Append(wx_pg.ColourProperty(str(i[0]), "", i[1]))

        self.m_palette_grid.SetSplitterPosition(50)

    def OnClose(self, event) -> None:
        wx.LogMessage("Closing the program.")

        # Executor will be terminated automatically because it is a daemon

        self.Destroy()

    def OnExit(self, event) -> None:
        self.Close()

    def OnAbout(self, event) -> None:
        info = Img2STLAboutDialog(self)

        info.set_icon(self.app_icon)
        info.set_name(ST_NAME, ST_VERSION)
        info.set_description(ST_DESCRIPTION)
        info.set_using("using " + wxbuildinfo())
        info.set_license(ST_LICENSE)
        info.set_copyright(ST_LICENSE_LINK, ST_COPYRIGHT)
        info.add_link("https://github.com/baskiton/Img2STL", "Homepage: GitHub")

        info.fit()
        info.ShowModal()
        info.Destroy()

    def main_containerOnSplitterSashPosChanging(self, event) -> None:
        self.Layout()

    def prop_image_openOnButtonClick(self, event) -> None:
        open_file_dlg = wx.FileDialog(
            self,
            "Open image files",
            self._files_directory,
            wx.EmptyString,
            f"Images file (*.bmp *.gif *.png *.jpg)|*.bmp;*.gif;*.png;*.jpg|"
            f"All files (*.*)|*.*",
            wx.FD_OPEN | wx.FD_MULTIPLE | wx.FD_CHANGE_DIR
        )

        if open_file_dlg.ShowModal() == wx.ID_CANCEL:
            return

        self._files_directory = open_file_dlg.GetDirectory()
        self._exit_files_dir = self._files_directory
        self._files_list = open_file_dlg.GetFilenames()
        self._exit_files = open_file_dlg.GetFilenames()

        names = "; ".join(self._exit_files)
        for idx, item in enumerate(self._exit_files):
            fn, _ = os.path.splitext(item)
            self._exit_files[idx] = f"{fn}.stl"

        wx.LogMessage(f"Files {names} opened.")

        self.set_prop_exit_label()
        self.add_image_tabs()

        self.m_execute_button.Enable(True)
        self.m_prop_exit_btn.Enable(True)

    def prop_image_autocropOnCheckBox(self, event) -> None:
        self.change_image_scale()

    def prop_mask_mode_valOnChoice(self, event) -> None:
        val = self.m_prop_mask_mode_val.GetSelection()

        self.m_prop_transp_col_lable.Enable(val)
        self.m_prop_transp_col_val.Enable(val)

        self.change_image_scale()

    def prop_transp_col_valOnButtonClick(self, event) -> None:
        col = wx.ColourData()
        col.SetColour(self.m_prop_transp_col_val.GetBackgroundColour())
        col_dlg = wx.ColourDialog(self, col)

        if col_dlg.ShowModal() == wx.ID_CANCEL:
            return

        col.SetColour(col_dlg.GetColourData().GetColour())
        self.m_prop_transp_col_val.SetBackgroundColour(col.GetColour())
        self.m_prop_transp_col_val.SetLabel(col.GetColour().GetAsString(wx.C2S_HTML_SYNTAX))

        self.change_image_scale()

    def prop_exit_btnOnButtonClick(self, event) -> None:
        count = len(self._files_list)

        if count == 1:      # single file
            save_file_dlg = wx.FileDialog(
                self,
                "Save File",
                self._exit_files_dir,
                self._exit_files[0],
                f"STL file (*.stl)|*.stl|"
                f"All files (*.*)|*.*",
                wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT
            )
            if save_file_dlg.ShowModal() == wx.ID_CANCEL:
                return
            self._exit_files_dir = save_file_dlg.GetDirectory()

        elif count > 1:     # multi files
            save_dir_dlg = wx.DirDialog(
                self,
                "Save to directory",
                self._exit_files_dir
            )
            if save_dir_dlg.ShowModal() == wx.ID_CANCEL:
                return
            self._exit_files_dir = save_dir_dlg.GetPath()

        wx.LogMessage(f"Exit directory changed to {self._exit_files_dir}")

        self.set_prop_exit_label()

    def execute_buttonOnButtonClick(self, event) -> None:
        self.m_progress_bar.SetValue(0)

        self._executor = Executor(
            evt_hdlr=self,
            autocrop=self.m_prop_image_autocrop.GetValue(),
            density=self.m_prop_den_val.GetValue(),
            height_max=self.m_prop_max_height_val.GetValue(),
            height_min=self.m_prop_min_height_val.GetValue(),
            mask_mode=self.m_prop_mask_mode_val.GetSelection(),
            mask_color=self.m_prop_transp_col_val.GetBackgroundColour(),
            height_mode=self.m_prop_height_mode_val.GetSelection(),
            f_type=self.m_prop_exit_format.GetSelection(),
            files_dir=self._files_directory,
            files_list=self._files_list,
            exit_files_dir=self._exit_files_dir,
            exit_files_list=self._exit_files
        )

        self._executor.start()

        self.m_prop_panel.Enable(False)
        self.m_progress_label.SetLabel("Process...")
        self.m_progress_label.CentreOnParent(wx.HORIZONTAL)

    def image_panel_containerOnSize(self, event: wx.SizeEvent) -> None:
        # FIXME
        # I don’t understand why yet, but this method is called twice,
        # so this is where the "protection" is organized using a counter

        if self._cnt:
            self._cnt = False
            self.m_image_panel_container.SetSashPosition(event.GetSize().GetWidth() - 200)
            self.change_image_scale()
        else:
            self._cnt = True

    def image_panel_containerOnSplitterSashPosChanged(self, event) -> None:
        self.m_image_info_panel.SetSize(200, -1)

    def noteOnAuiNotebookPageChanged(self, event) -> None:
        # TODO
        # if self.m_note.GetClientData():
        #     return

        self.change_image_scale()

    def noteOnAuiNotebookPageClose(self, event) -> None:
        if self.m_note.GetPageCount() <= 0:
            self.m_palette_grid.Clear()
            self.m_palette_grid.Append(wx_pg.PropertyCategory("amount | color"))



from .utils import get_bound_box, img_to_heightmap, ThreadEvent, WorkerMessage
from .logging import Logging
from .cmd import ThrCmds, WMCmds
from .executor import Executor
from .worker import Worker

__all__ = (
    'get_bound_box',
    'img_to_heightmap',
    'ThreadEvent',
    'WorkerMessage',
    'Logging',
    'ThrCmds',
    'WMCmds',
    'Executor',
    'Worker',
)
